// tin.cc
// Time-stamp: </home/wrf/c/2019/tin/tin.cc, Wed, 20 Feb 2019, 06:29:48 EST, w540, wrf, W. Randolph Franklin, https://wrf.ecse.rpi.edu/>


/*
W. Randolph Franklin, Professor
ECSE Dept., 6026 JEC,
Rensselaer Polytechnic Inst,
110 8th St,
Troy NY, 12180 USA

mail@wrfranklin.org or frankwr@rpi.edu 
+1 (518) 276-6077 

https://wrf.ecse.rpi.edu/
*/
  
// Time-stamp: </wrf/c/tin/tin.cc, Tue, 20 May 2003, 11:42:50 EDT,
// wrf@benvolio.ecse.rpi.edu>



// INCLUDES 


#include <fcntl.h>
#include <fstream>
#include <iomanip>
#include <iostream>
#include <malloc.h>
#include <math.h>
#include <ostream>
#include <sstream>
#include <stdio.h>  // for perror
#include <stdlib.h>
#include <streambuf>
#include <string>
#include <sys/file.h>
#include <sys/times.h>
#include <sys/types.h>
#include <sys/uio.h>
#include <unistd.h>

using namespace std;



/* BUILTIN FUNCTION DECLARATIONS */

//extern off_t lseek (int fd, off_t offset, int whence);
//extern int sscanf ();



// NECESSARY FORWARD DECLARATIONS

struct Triangle;
struct Point;
struct Heap;
struct Linked_list;
class Dotpair;
template <class T> inline void  swapvar(T &a, T &b);



// TYPEDEFS

typedef float COORDT;



// CONSTANTS

const float TOLERANCE = 1.0e-6;	// Numbers Smaller Than This Are Worrisome. 
const int MFLIPQ = 100;		/* Size of queue of edges to consider for flipping.
				   This should be a little bigger than you'd think
				   since edges are appended to the end while they
				   are being processed from the front. */
const int MDELETABLE = 3;	/* Size of array of deletable, null, triangles. Need
				   never be over 3. */
const int plus1[3] = { 1, 2, 0 };	// Plus 1, Modulo 3 
const int minus1[3] = { 2, 0, 1 };	// Minus 1, Modulo 3 




// VARIABLES

int tinn;		// Number Of Rows, Columns Of Data 
int flipq[MFLIPQ][2];	/* Queue of edges to consider to flipping. Each
				   entry is given as the IDs of the 2 adjacent
				   triangles, since the edges themselves are not
				   stored explicitly.  */
int nflipq;		// # edges in q now. 
int hiflipq;		// Hi-water size for edge flip q.
int nflips;		// # flips made so far. 
int deletable_triangle[MDELETABLE];	/* Flagged for being null and
						   exterior, but not yet deleted. */
int ndeletable;		// Size Of Contents Of Deletable_Triangle. 
float   goal_err;		// Stop splitting when max_err_over_whole_TIN is
				// this small.
int tsplits;		// Total Number Of Splits To Do. 
int     tgoalerr;		// Stop splitting at this error; scaled up by 1000
				// to make it an int for easy input.
int tflip;		// When to flip quadrilateral diagonals.
int tincrfiles;		// Whether, and how often, to write incremental edge and triangle
				// files.
float   clocks_per_sec = 0.0;	// will be set first thing in main.
int     twritemore;		// Write more output: ascii triangles, edges,
				// centers, pgm file? 
int     tstats;			// How often to write a line to stats file? Do it
				// every tstats splits.


Point *points;		/* Array of points: (x,y,z).  Heights are already scaled. X and y are explicit to allow for future generalization. */
int     npoints;
Triangle *triangles;
int     ntriangles;		// Current number of triangles
int     max_allowable_num_triangles;	// Max allowable number of triangles
float   max_err_over_whole_TIN;	// Maximum error for whole TIN.
int     worst_triangle;		// ID of worst triangle, with the maximum error. 
int     nsplits;		// # splits so far. 
int     tprint;			// Amount Of Printing To Do. 0 - No Routine
				// Printing. 

// Statistics

int n_elts_pushed_onto_heap;	// Number of times elements pushed onto
					// errs.
int n_elts_popped_from_heap;	// Number of times elements popped from
					// errs.
int n_elts_deleted_from_heap;	// Number of times elements deleted from
					// errs.
int n_swaps_while_pushing_heap;	// Number of times elements swapped pushing.
int n_swaps_while_popping_heap;	// Number of times elements swapped popping.
int max_ever_heap_size;	// Hi-water mark for heap size.
int n_elts_pushed_onto_errs_heap;		// Number of times elements pushed onto errs.
int n_elts_pushed_onto_errs_heap_swap;		// Number of times elements swapped pushing.
int n_split_tri_list_traverses;
int n_new_traverses;

ofstream statfile ("stats");	// Each line: # splits, # triangles,
				// max_err_over_whole_TIN, # flips.

ofstream incplotfile("incplot.gnu");   // gnuplot command file to plot incremental triangulations

int nincplots;   // Number of incremental plots on a side, in the multiplot.
float incplotsize;  // Size of one plot in the multiplot.


Dotpair *dotpairs;		// Array of dotted pairs for linked lists of points
				// in triangles.  

Heap   *errs;			// Heap of errors of each triangle.



// ROUTINE DECLARATIONS NEEDED FOR CLASS DECLARATIONS

void    Test_Tri (const int t, const string msg);
void    Test_Pt (const int t, const string msg);
void    perror2 (const string);




// LINKED LIST CLASSES.  
// All linked lists are allocced from one global array, pairs.

class   Linked_list;
typedef int Pairindex;
const Pairindex eolist (-1);	// End of dotted pair list. 

class   Dotpair			// One dotted pair, containing a point id and an
				// index to the
{				// next dotted pair.
public:
  int     ptid;		// ID of a point, i.e., index in points array. 
  Pairindex next;		// The remaining points.
};


class   Linked_list {
public:
  Pairindex head;		// Index in dotpairs of the 1st element.

  // ==eolist if linked list is empty.
  Linked_list () {
    head = eolist;
  } 

  Linked_list (const int a) {
    head = a;
  }				// ???

  void    push (int point);	// Prefix a new point to this list

  int     pop ();		// Pop the first point off the list.

  void    check () const	// Check that there is not a loop at the start.
  {
    if (head != eolist && head == dotpairs[head].next)
      perror2 ("Checking list found that the first pointer loops.");
  }

  int     length () const		// Return number of elements in list.
  {				// Cannot easily be recursive while catching
				// infinite loops.
    int     i = 0;
    Pairindex l = head;
    while (l != eolist) {
      l = dotpairs[l].next;
      i++;
      if (i > npoints)
	perror2 ("Apparent infinite loop in Linked_list.length.");
    }
    return i;
  }
};


Linked_list freelist;		// Free list of unallocated dotted pairs in
				// dotpairs.



struct Point			// One point.  A struct is 4 bytes shorter than an
				// array.
{
  short int x, y, z;
};


class Triangle {
public:
  int     vert[3];		// ID Numbers of the 3 verts.  vert[i] is opposite
  // adj[i].  The vertices around a triangle are ordered in
  // the positive direction.  Vertices are not also stored
  // in the list of points in this triangle.
  int     adj[3];		// ID Numbers of the 3 adjacent triangles, or -1 if
                                // that side is exterior.
  int     heap_loc;		// Location in heap of this triangle.  -1 if not in
				// heap.
  int     worst_pt;		// ID of the worst point.  This is its index in the
                                // global points array.  
  Linked_list l;		// Linked list of the ID numbers of the points
				// inside this triangle.
  float   maxe;		// Its error from the plane of the triangle,
  // measured vertically.  Type is not short int ob possible
  // overflow.

  Triangle & operator = (const Triangle & t)	// Copy triangle but don't
    // update heap.
  {
    if (this == &t)
      return *this;
    this->vert[0] = t.vert[0];
    this->vert[1] = t.vert[1];
    this->vert[2] = t.vert[2];
    this->adj[0] = t.adj[0];
    this->adj[1] = t.adj[1];
    this->adj[2] = t.adj[2];
    this->heap_loc = t.heap_loc;
    // if (a.heap_loc >= 0) 
    // errs.elts[a.heap_loc].tri = (this - &triangles[0])/sizeof(Triangle);

    this->worst_pt = t.worst_pt;
    this->l = t.l;
    this->maxe = t.maxe;
    return *this;
  }};



// HEAP to store pairs of (float,int) representing a triangle error and number,
// and to return and delete the triangle with the largest error, i.e., the
// triangle listed in the pair with the largest first element.  The 1st
// element of the heap is the largest.  Each element is at least as large
// as its sons.  The sons of element #i are in locs #2i and #2i+1.  To make
// the indexing simpler, the 0th element is not used.  The interface is via
// the triangle number.

// When a triangle is split, deleted, or flipped, delete its entry here.

// Each triangle has a pointer to its entry in the heap (or -1 if the
// triangle is not in the heap).  This pointer must be updated when
// the heap entry is moved.

struct Heap_elt {
  int     tri;
  float   err () {
    return triangles[this->tri].maxe;
  }};

ostream & operator << (ostream & s, Heap_elt e);


int     Heap_Elt_Compare (const void *a, const void *b);


void swap_heap_elt (Heap_elt & a, Heap_elt & b)
{
  Test_Tri (a.tri, "1st arg to heap_elt swap");
  Test_Tri (b.tri, "2nd arg to heap_elt swap");
  swapvar (a.tri, b.tri);
  swapvar (triangles[a.tri].heap_loc, triangles[b.tri].heap_loc);
}


class   Heap {
public:
  int     alloc_size;		// Allocated size, excluding 0th element.
  int     heap_size;		// Current size
  Heap_elt *elts;

  Heap (const int size) {
    elts = new Heap_elt[size + 1];
    if (elts == 0)
      perror2 ("Cannot alloc elts in heap.");
    heap_size = 0;
    alloc_size = size;
  } void  push (const int t);

  // Pop triangle with largest err, setting global vars worst_triangle and
  // max_err_over_whole_TIN.  If no triangle, set -1 and 0.0 .

  void    pop ();

  // Delete entry #i of heap.

  void    del (const int i);
};




// ENV OPTIONS STUFF

typedef struct			// Type to define numerical envar options 
{
  string  name;		// envar 
  int     defaul;		// default.  Note 'default' is a reserved word. 
  int     minval, maxval;	// allowable range 
  int    &var;		// variable to store value into. 
} numerical_opt_t;

const numerical_opt_t numerical_opts[] = { {"TINN", 1, 10, 1201, tinn},
					   {"TGOALERR", 0, 0, 1000000000, tgoalerr},
					   {"TINCRFILES", 0, 0, 1000000, tincrfiles},
					   {"TPRINT", 0, 0, 1, tprint},
					   {"TSPLITS", 10000, 0, 1000000000, tsplits},
					   {"TSTATS", 10, 0, 10000000, tstats},
					   {"TFLIP", 1, 0, 4, tflip},
					   {"TWRITEMORE", 0, 0, 1, twritemore}
};

const int n_numerical_opts = 8;




// ROUTINE DECLARATIONS 

float   Area (const Triangle t);
double  Read_Delta_Time (), Read_Delta_Time2 ();
int     Split_Triangle ();
void    Add_to_Flipq (int t, int a);
void    Alloc_Dotpairs (const int n);
void    Calc_Triangle_Error (Triangle&);
void    Calc_Plane_Eqn (const Triangle&, float *const eqn);
void    Copy_Triangle (int to, int from);
void    Delete_Flagged_Triangles ();
void    Flag_Null_Triangle_For_Deletion (int i);
void    Get_Options ();
void    Init_Vars ();
void    Make_First_2_Triangles ();
void    Print_Help ();
void    Print_Points ();
void    Print_Stats ();
void    Print_Time (string), Print_Time2 (string);
void    Print_Triangle (int);
void    Print_Triangle2(const Triangle &);
void    Proc_Flipq ();
void    Find_Angle_Cosines(COORDT oldedge[2][3][2], COORDT newedge[2][3][2],
			float &oldmaxcos, float &newmaxcos) ;
void    Flip_Diagonal(int, int, Triangle*, COORDT*, int*, int);
int     Is_Flip_Possible(COORDT newedge[2][3][2]);
void    Read_Data ();
void    Update_Adjacency (int, int);
void    Write_Edges (const string);
void    Write_Pgm ();
void    Write_Triangles (const string);
void    Write_Triangle_Centers (const string);
void    Write_Plotfile(const string plotfilename, const string centerfilename,
		    const string edgefilename);
void    Print_Split_Status(const int nsplits);
void    Write_Incr_Output(const int nsplits);



// INLINE FUNCTIONS

/* ABS */

template < class T > inline T abs (const T a)
{
  return (a >= 0) ? a : -a;
}


#ifdef OLD
/* MAX */
template < class T > inline T max (const T x, const T y)
{
  return (x > y) ? x : y;
}

/* MIN */
template < class T > inline T min (const T x, const T y)
{
  return (x < y) ? x : y;
}
#endif

/* SWAPVAR Swap 2 variables */

template <class T>
inline void 
swapvar(T &a, T &b)
{
  T     c;
  c = a;
  a = b;
  b = c;
}



// DOT2D - Dot Product Of 2 2-D Vectors 

template <class T>
inline T Dot2D (const T *const a, const T *const b)
{
  return (a)[0] * (b)[0] + (a)[1] * (b)[1];
}


// DOT3D - DOT_PRODUCT In 3-D 

template < class TA, class TB >
inline float Dot3D (const TA * const a, const TB * const b)
{
  return a[0] * b[0] + a[1] * b[1] + a[2] * b[2];
}

template <> inline float Dot3D (const Point * const a, const float *const b)
{
  return a->x * b[0] + a->y * b[1] + a->z * b[2];
}


// DET2D - Determinant Of 2 2-D Vectors 

template <class T>
inline T Det2D (const T *const a, const T *const b)
{
  return (a)[0] * (b)[1] - (a)[1] * (b)[0];
}


// DET3 - Determinant Of 3 Edges

template <class T>
inline T Det3 (T a[3][2])
{
  return a[0][0]*a[1][1] + a[0][1]*a[2][0] + a[1][0]*a[2][1]
    - a[0][0]*a[2][1] - a[0][1]*a[1][0] - a[1][1]*a[2][0];
}


// MINEQ - Set the first arg to the min of the 2 args, and return it.

inline float Mineq (float &a, const float b)
{
  if (b < a)
    a = b;
  return a;
}

// SQUARE - Convert its COORDT arg to float and square it.

template <class T>
inline float Square (const T k)
{
  return static_cast<float> (k)  * static_cast<float> (k);
}






// ================================================================


// PROCESS_CPU_TIME Return CPU (user+system) time since start of process.

double Process_CPU_Time ()
{
  struct tms *time_buffer = new tms;
  (void) times (time_buffer);
  return ((time_buffer->tms_utime + time_buffer->tms_stime) / clocks_per_sec);
}



/**; READ_DELTA_TIME

* Returns time in seconds since last Read_Delta_Time.  Automatically
* initializes itself on 1st call and returns 0.  Also, return time
* from process start in its arg.
*/

double Read_Delta_Time (double &new_time)
{
  static double old_time = 0.0;
  double  delta;

  new_time = Process_CPU_Time ();
  delta = new_time - old_time;
  old_time = new_time;
  return delta;
}

/**; PRINT_TIME Print time since last call, with a message. */

void Print_Time (string msg)
{
  double  tottime;
  double  incrtime = Read_Delta_Time (tottime);
  cout << "Total CPU Time thru " << msg << " =" << tottime << ", increment="
       << incrtime << endl;
}

/**; READ_DELTA_TIME2
 * Like Read_Delta_Time but keeps a second incremental count.
 */

double Read_Delta_Time2 ()
{
  static struct tms *time_buffer = NULL;
  static time_t old_time2, new_time2;

  double  delta;

  if (time_buffer == NULL) {
    time_buffer = (struct tms *) malloc ((unsigned) sizeof (struct tms));
    old_time2 = 0;
  }
  (void) times (time_buffer);
  new_time2 = time_buffer->tms_utime + time_buffer->tms_stime;
  delta = (new_time2 - old_time2) / clocks_per_sec;
  old_time2 = new_time2;
  return delta;
}

/**; PRINT_TIME2 Print time since last read_delta_time2 call, with a message. */

void Print_Time2 (string msg)
{
  cout << "CPU Time for " << msg << " =" << Read_Delta_Time2 () << endl;
}



// PERROR2 Print fatal error message, then statistics.

void perror2 (const string msg)
{
  static int loop = 0;

  perror (msg.c_str());
  if (++loop > 1) {
    cout << "Infinite loop in perror2!" << endl;
    exit (15);
  }
  Print_Stats ();
  exit (11);
}




// MAIN 

main ()
{
  char    split_label[100];

  clocks_per_sec = sysconf (_SC_CLK_TCK);
  cout << endl
       << "****************************************************************"
       << endl
       <<
    "[TIN - Triangular Irregular Network program, Wm. Randolph Franklin, 1973 -  May 2003]"
       << endl;
  cout << "[Compiled from " << __FILE__ << " on " << __DATE__ << ", " << __TIME__
       << ']' << endl;
  cout << "Run date and environment: " << endl;
  system ("/bin/date");
  system ("/bin/uname -a");
  cout << endl;

  if (!statfile)
    perror2 ("Can't open stats for writing.");

  if (!incplotfile)
    perror2 ("Failed to open incplotfile for writing.");

  Get_Options ();
  Init_Vars ();
  Read_Data ();
  Read_Delta_Time2 ();
  Make_First_2_Triangles ();
  errs->pop ();		// Get worst triangle.


  // Loop, splitting the worst triangle.  Stop when the any one of
  // several terminating conditions becomes true: after a predefined
  // number of splits, because no more triangles are splittable, or when
  // the max error falls below a certain value.  At the start of each
  // iteration, the worst triangle has already been popped off the heap,
  // and is stored in worst_triangle and max_err_over_whole_TIN.

  statfile << "# Statistics file for tin:" << endl
	   << "# One line written after each triangle split." << endl
	   <<
    "# Each line: Number of triangles split so far, current number of triangles,"
	   << endl <<
    "#  max error over whole TIN, number of edges flipped so far, cumulative CPU time."
	   << endl;


  nincplots=1;   // Number of incremental plots on a side, in the multiplot.
  if (tincrfiles>0) 
  nincplots 
    = static_cast<int>(sqrt(static_cast<float>(tsplits)/static_cast<float>(tincrfiles))+0.999999);
  if (tincrfiles==99) nincplots=5;
  incplotsize= 1.0/static_cast<float>(nincplots);

  if (tprint>0) 
    cout << "nincplots=" << nincplots << ", incplotsize=" << incplotsize << endl;

  incplotfile << "#!/usr/bin/gnuplot\n"
	      << "set xlabel\n"
	      << "set ylabel\n"
	      << "set noxtics\n"
	      << "set noytics\n"
	      << "set nokey\n"
	      << "set multiplot\n" 
	      << "set size " << (incplotsize*0.9) << ',' << (incplotsize*0.9) << '\n'
	      << "set origin 0,0\n";

  for (nsplits = 1; nsplits <= tsplits; nsplits++) {
    if (max_err_over_whole_TIN <= goal_err) {
      cout << "Stopping because max_err_over_whole_TIN="
	   << max_err_over_whole_TIN << " <= goal_err=" << goal_err << endl;
      break;
    }
    if (Split_Triangle () != 0) {
      cout << "Stopping because no triangle is splittable." << endl;
      break;
    }
    Proc_Flipq ();
    errs->pop ();		// Get the new worst triangle.
    if (tprint != 0)
      cout << "After " << nsplits << " splits, ntriangles= " << ntriangles
	   << ", max_err_over_whole_TIN=" << max_err_over_whole_TIN
	   << ", nflips=" << nflips << endl;
    if (tstats >0 && nsplits % tstats == 0)
      statfile << setw (8) << nsplits << setw (8) << ntriangles << setw (9)
	       << max_err_over_whole_TIN << setw (8) << nflips
	       << setw (9) << Process_CPU_Time () << endl;
    if ((nsplits % 10000) == 0) Print_Split_Status(nsplits);

    if (tincrfiles==99) {
      // The following mess calls Write_Incr_Output every time nsplits
      // has increased by 20%.
      float a1=log(nsplits)/log(1.414);
      float a2=log(nsplits+1)/log(1.414);
      //      cout << "nsplits, a1, a2= " << nsplits << ',' << a1 << ',' << a2 << endl;
      if (floor(a1)<floor(a2)) 
	Write_Incr_Output(nsplits);
    }   // this brace controls which 'if' the following 'else' attaches to.
    else if (tincrfiles>0) 
      if (nsplits % tincrfiles == 0) Write_Incr_Output(nsplits);
  }
  if (nsplits >= tsplits)
      cout << "Stopping because " << nsplits << " splits performed." << endl;
  if (tincrfiles>0) 
    incplotfile << "set nomultiplot\n"
		<< "pause -1" << endl;
  Print_Stats ();
}


// WRITE_INCR_OUTPUT Write the edges.NNN, centers.NNN, plot.NNN.gnu
// files showing how the program's operation.  Also append to the
// incremental plot file.  This gives a choice of seeing one
// incremental plot, or all of them at once.  If plotting them all at
// once, or if there are more than 100 triangles, then don't plot
// triangle numbers or write triangle centers; it's too busy.

void Write_Incr_Output(const int nsplits)
{
  ostringstream edgestring("edges_",ios::ate);
  edgestring << nsplits << ".dat";
  //  cout << "Edge file is >>" << edgestring.str() << "<<" << endl;
  Write_Edges (edgestring.str());

  ostringstream plotstring("plot.",ios::ate);
  plotstring << nsplits << ".gnu";
  if (nsplits<100) {
    ostringstream centerstring("centers.",ios::ate);
    centerstring << nsplits;
    Write_Triangle_Centers (centerstring.str());

    Write_Plotfile (plotstring.str(), centerstring.str(), edgestring.str());
  }
  else
    Write_Plotfile (plotstring.str(), string(), edgestring.str());

  float xorigin = incplotsize * ((nsplits-1)/nincplots);
  float yorigin = incplotsize * ((nsplits-1)%nincplots);

  incplotfile << "set origin " << xorigin << ',' << yorigin << '\n'
	      << "set size " << (incplotsize*0.9) << ',' << (incplotsize*0.9) 
	      << '\n'
    //	      << "load \"" << centerstring.str() << '"' << '\n'
	      << "plot \"" << edgestring.str() << "\" with lines" << endl;
}



// PRINT_SPLIT_STATUS Lets me know how the program is progressing, as
// it is running.

void Print_Split_Status(const int nsplits)
 {
   ostringstream csplit;
   n_new_traverses = n_split_tri_list_traverses - n_new_traverses;
   // cout << n_new_traverses << ' ';
   n_new_traverses = n_split_tri_list_traverses;
   csplit << "after " << nsplits << " splits";
   Print_Time (csplit.str()); 
 }



// *; GET_OPTIONS Get Input Options From Envariables 

void Get_Options ()
{
  char   *s;
  int     i;

  const numerical_opt_t *thiss;

  for (i = 0; i < n_numerical_opts; i++) {
    thiss = &numerical_opts[i];
    s = getenv ((thiss->name).c_str());
    if (s == NULL)
      thiss->var = thiss->defaul;
    else {
      (void) sscanf (s, "%d", &(thiss->var));
      if (thiss->var < thiss->minval || thiss->var > thiss->maxval) {
	cout << "Error: illegal value " << thiss->var << " for" << thiss->
	  name << ", legal range=(" << thiss->minval << ',' << thiss->
	  maxval << ").\n";
	thiss->var = thiss->defaul;
      }
    }
  }

  goal_err = tgoalerr * 0.001;

  if (tinn <= 1)
    Print_Help ();

  cout << "Input params: TINN=" << tinn << ", TPRINT=" << tprint
       << ", TINCRFILES=" << tincrfiles << ", TGOALERR=" << tgoalerr
       << ", TFLIP=" << tflip << ", TSPLITS=" << tsplits 
       << ", TSTATS=" << tstats << ", TWRITEMORE=" << twritemore << endl;
  Print_Time ("Get_Options");
}



/* 
   INIT_VARS Better here than in an initialization in a declaration since now I can
   restart the program w/o reloading it. */

void Init_Vars ()
{
  long int tinn2 = tinn * (long int) tinn;
  short int *p;
  int     i;

  nflipq = 0;
  nflips = 0;
  hiflipq = 0;
  max_ever_heap_size = 0;
  ndeletable = 0;
  max_allowable_num_triangles = 2 * tsplits + 5;	// Slightly conservative (real max is 2*tsplits+2)

  if (0 == (points = new Point[tinn2]))
    perror2 ("Allocating points array in Init_Vars.");
  if (0 == (triangles = new Triangle[max_allowable_num_triangles]))
    perror2 ("Allocating triangles array in Init_Vars.");

  Alloc_Dotpairs (tinn2*2);
  if (0 == (errs = new Heap (tsplits * 2)))
    perror2 ("Allocating error heap in Init_Vars");
  n_elts_pushed_onto_errs_heap = 0;
  n_elts_popped_from_heap = 0;
  n_elts_deleted_from_heap = 0;
  n_elts_pushed_onto_errs_heap_swap = 0;
  n_swaps_while_popping_heap = 0;
  n_split_tri_list_traverses = 0;
  n_new_traverses = 0;
  Print_Time ("Init_Vars");

}



// ALLOC_dotpairs Allocate the dotted pairs array and link the freestore.

void Alloc_Dotpairs (const int n)
{
  int     i;
  dotpairs = new Dotpair[n];
  if (dotpairs == 0)
    perror2 ("Unable to alloc dotpairs in alloc_dotpairs.");
  freelist.head = 0;
  for (i = 1; i < n; i++)
    dotpairs[i - 1].next = i;
  dotpairs[n - 1].next = eolist;
  for (i = 0; i < n; i++)
    dotpairs[i].ptid = -9999;
}



/** READ_DATA

1. Open stdin containing a square array of short int heights.
2. Do some validity checks on constants.
3. Read the heights into a linear array, supply Xs and Ys, and
determine bounds.
*/

void Read_Data ()
{
  int     i, j;
  short int zmin, zmax;

  zmin = 32767;
  zmax = -32767;

  // Copy The Heights Data To The Points Array. 

  npoints = 0;
  for (i = 0; i < tinn; i++)
    for (j = 0; j < tinn; j++) {
      points[npoints].x = i;
      points[npoints].y = j;
      cin >> points[npoints].z;
      if (!cin.good ()) {
	cout << "At npoints=" << npoints << endl;
	perror2 ("Reading cin is not good.");
      }

      zmin = min (zmin, points[npoints].z);
      zmax = max (zmax, points[npoints].z);
      npoints++;
    }
  cout << "Z range is [" << zmin << "," << zmax << "]\n";

  Print_Time ("Read_Data");

#ifdef DEBUG
  Print_Points ();
#endif
}



/* 
   CALC_TRIANGLE_ERROR Calculate the plane equation of a triangle,
   and then find its maximum error, both which point and the value.
   Store the worst point and its error in the triangle.  The plane
   equation is not stored in the triangle since we never need it again
   after finding worst_pt and worst_err.
*/

void Calc_Triangle_Error (Triangle &t)
{
  int worst_pt = -1;
  COORDT worst_err = 0;

  /* eqn: the equation of the plane defined by the triangle's 3
     vertices, in the form z=x*eqn[0]+y*eqn[1]+eqn[2].  */

  float   eqn[3];	

  Calc_Plane_Eqn (t, eqn);

  Pairindex thispair = t.l.head;
  while (thispair != eolist) {
    if (thispair < 0)
      perror2 ("Negative thispair in Calc_Triangle_Error");
    Point &pt = points[dotpairs[thispair].ptid];
    COORDT d1 = fabs (eqn[0] * pt.x + eqn[1] * pt.y + eqn[2] - pt.z);
    if (d1 > worst_err) {
      worst_err = d1;
      worst_pt = dotpairs[thispair].ptid;
    }
    thispair = dotpairs[thispair].next;
  }
  t.worst_pt = worst_pt;
  t.maxe = worst_err;
}




// CROSS_PRODUCT A = B X C 

template < class TA, class TB, class TC >
void Cross_Product (TA * const a, const TB * const b, const TC * const c)
{
  a[0] =
    static_cast < TA > (b[1]) * static_cast < TA > (c[2]) - static_cast < TA >
    (b[2]) * static_cast < TA > (c[1]);
  a[1] =
    static_cast < TA > (b[2]) * static_cast < TA > (c[0]) - static_cast < TA >
    (b[0]) * static_cast < TA > (c[2]);
  a[2] =
    static_cast < TA > (b[0]) * static_cast < TA > (c[1]) - static_cast < TA >
    (b[1]) * static_cast < TA > (c[0]);
}




/* 
   CALC_PLANE_EQN Calculate equation of a triangle from its vertices.  Its form is z=ax+by+c. */

void Calc_Plane_Eqn (const Triangle &t, float *const eqn)
{
  Point  *a, *b, *c;		// The Vertices' Addresses (Not Their Ids). 
  int     db[3], dc[3];	// The Delta Vectors For 2 Sides. 
  float   u[3];		// Unnormalized Normal Vector. 
  float   d;			// When Eqn Is In Form Ax+By+Cz=D, This Is 'D'. 
  int     i;

  a = &points[t.vert[0]];
  b = &points[t.vert[1]];
  c = &points[t.vert[2]];

  db[0] = b->x - a->x;
  db[1] = b->y - a->y;
  db[2] = b->z - a->z;

  dc[0] = c->x - a->x;
  dc[1] = c->y - a->y;
  dc[2] = c->z - a->z;

  Cross_Product (u, db, dc);

  if (abs (u[2]) < TOLERANCE) {
    if (tprint != 0) {
      cout << "Warning: Plane of this triangle is nearly vertical:" << endl;
      Print_Triangle2(t);
    }
    eqn[0] = 0.0;
    eqn[1] = 0.0;
    eqn[2] = 0.0;
  } else {
    eqn[0] = -u[0] / u[2];
    eqn[1] = -u[1] / u[2];
    d = Dot3D (a, u);
    eqn[2] = d / u[2];
  }
}




// WRITE_POINT  Write a point, both as a number and coords.

void Write_Point(const int p)
{
  cout << p << '(' << points[p].x << ',' << points[p].y << ')';
}




// PRINT_TRIANGLE 

void Print_Triangle (const int t)
{
  Triangle &tt = triangles[t];

  if (tprint == 0)
    return;

  cout << "Triangle #" << t << ": vert: ";
  Print_Triangle2(tt);
}

void Print_Triangle2(const Triangle &tt)
{
  Write_Point(tt.vert[0]);
  cout << ' ';
  Write_Point(tt.vert[1]);
  cout << ' ';
  Write_Point(tt.vert[2]);
  cout << ", adj: " << tt.adj[0] << ' ' << tt.adj[1] << ' ' << tt.adj[2]
       << "," << endl << "   max_err_over_whole_TIN=" << tt.maxe << " at " 
       << tt.worst_pt << ", np=" << tt.l.length () << "\n";
#ifdef DEBUG
  cout << "   p=";
  for (int i = 0; i < tt.np; i++) {
    cout << ' ' <<  tt.p[i];
    if ((i && 15) == 0 && i > 0)
      cout << endl << "   ";
  }
  if ((i == 0) || ((i && 15) != 0))
    cout << endl;
  cout << "worst_pt=" << tt.worst_pt << ", maxe=" << tt.maxe;
#endif
}


// PRINT_POINTS All Of Them 

void Print_Points ()
{
  cout << "Npoints=" << npoints;

  for (int i = 0; i < npoints; i++) {
    cout << i << ": " << points[i].x << ", " << points[i].y << ", " << points[i].
      z;
    if ((i & 1) == 1)
      cout << '\n';
  }
  cout << endl;
}



// FLAG_NULL_TRIANGLE_FOR_DELETION: Test whether triangle has no
// interior points, has zero area, and is also on the outside of the
// square.  If so, flag for later deletion. It's too messy to delete
// them now; that would interfere with putting them into the flip
// queue.  The deletable triangles are stored in an array.

// A null triangle that is not on the outside should be removed soon by
// having a diagonal flipped.

void Flag_Null_Triangle_For_Deletion (int t)
{
  int     i, j, k;
  Triangle & tri = triangles[t];

  if (tri.l.head != eolist)
    return;
  if (tri.adj[0] >= 0 && tri.adj[1] >= 0 && tri.adj[2] >= 0)
    return;
  if (abs (Area (tri)) > TOLERANCE)
    return;

  if (tprint != 0)
    cout << "Flagging triangle #" << t 
	 << " for deletion since it has no points, has zero area, and is on the outside." << endl;

  if (ndeletable >= MDELETABLE) {
    cout << "ERROR: ndeletable too large." << endl;
    exit (1);
  }
  deletable_triangle[ndeletable++] = t;
}



// DELETE_FLAGGED_TRIANGLES 

void Delete_Flagged_Triangles ()
{
  int     i, j, t;
  Triangle *thist;

  for (i = 0; i < ndeletable; i++) {
    t = deletable_triangle[i];
    thist = &triangles[t];
    if ((thist->l).head != eolist)
      cout << "Warning, in Delete_Flagged_Triangles, deleting triangle #" << t
	   << ", but it has a nonnull point list.\n";

    /* Any triangle that was adjacent to this triangle is now adjacent to the
       outside. */

    Update_Adjacency (t, -1);

    if (t < ntriangles - 1) {	// Copy Last Triangle Into This Position. 
      Copy_Triangle (t, ntriangles - 1);

      /* Any triangle that was adjacent to #ntriangles-1 is now adjacent to
	 #t. */
      Update_Adjacency (ntriangles - 1, t);
    }
    ntriangles--;
    if (tprint > 0)
      cout << "Deleting triangle #" << t << ", nt now=" << ntriangles << '\n';
  }
  ndeletable = 0;
}



/* 
   UPDATE_ADJACENCY The triangle in triangles[oldt] has been moved to
   triangles[newt].  Now update all references to it.  For every
   triangle that was adjacent to #oldt, make it adjacent to #newt in
   that entry.  Also change every instance of #oldt to #newt in flipq.
   If newt= -1 don't update flipq since there should be no oldt's in
   flipq. */

void Update_Adjacency (const int oldt, const int newt)
{
  int     i, j, t;

  for (i = 0; i < 3; i++) {
    t = triangles[oldt].adj[i];
    if (t >= 0) {		// Triangle T Is Adjacent.  Update It. 
      for (j = 0; j < 3; j++)
	if (triangles[t].adj[j] == oldt) {
	  triangles[t].adj[j] = newt;
#ifdef DEBUG
	  cout << "Updating adjacency of triangle " << t << " from " << oldt 
	       << " to " <<  newt << endl;;
#endif
	  goto foundadj;
	}
      cout << "ERROR: In Update_Adjacency, triangle #" << oldt 
	   << " is adjacent to triangle #" << t << "but the reverse is not true." 
	   << endl;
    foundadj:;
    }
  }

  if (newt >= 0)
    for (i = 0; i < nflipq; i++)
      for (j = 0; j < 2; j++)
	if (flipq[i][j] == oldt) {
	  flipq[i][j] = newt;
	  if (tprint != 0)
	    cout << "In update_adjacencies, flipq[" << i << "][" << j << "]=oldt=" 
		 << oldt << " changed to " << newt << "." << endl;
	}
}


/* 
   ADD_TO_FLIPQ Add a triangle pair, defining a new edge, and a new
   diagonal of a quadrilateral, to the flip queue, to be considered
   for flipping.  The inputs are one triangle number and the number of
   its adj entry defining the other triangle.  Don't add an edge if
   it's already in the queue. */

void Add_to_Flipq (const int t, const int a)
{
  int     i;

  if (tflip == 0)
    return;			// Maybe user doesn't want any flips.

  i = triangles[t].adj[a];
  if (i < 0)
    return;			// Neighbor Is The Outside. 
  if (nflipq >= MFLIPQ) {
    cout << "ERROR: MFLIPQ=" << MFLIPQ << " is too small.\n";
    exit (1);
  }
  flipq[nflipq][0] = min (t, i);
  flipq[nflipq][1] = max (t, i);

  // Is This Edge Already There? 

  for (i = 0; i < nflipq; i++)
    if (flipq[i][0] == flipq[nflipq][0] && flipq[i][1] == flipq[nflipq][1]) {
#ifdef DEBUG
      printf ("Not adding (%d,%d) to flipq since it duplicates entry #%d\n",
	      flipq[nflipq][0], flipq[nflipq][1], i);
#endif
      goto dupflip;
    }
#ifdef DEBUG
  printf ("Adding (%d,%d) to flipq[%d].\n", flipq[nflipq][0], flipq[nflipq][1],
	  nflipq);
#endif
  nflipq++;
  hiflipq = max (hiflipq, nflipq);
 dupflip:;
}


/* 
   WRITE_PGM Write the data about which triangle each point is in as a portable
   greymap (PGM) file so that it can be displayed for debugging purposes, by, e.g.,
   xv, which can pseudocolor the greyscales.  Write_Pgm works only when the points
   are in an array.  Assign the triangle numbers as their colors. Make the vertices
   white, coded by ntriangles in c. */

void Write_Pgm ()
{
  int     i, j, k, itr, x, y;
  int     c[tinn][tinn];	// will this work? 
  FILE   *f;


////////////////////////////////////////////////////////////////
/////////// pgm file is less useful than I thought
  return;
////////////////////////////////////////////////////////////////

  /* Zero c, which shouldn't be necessary since every entry should be set later,
     but be careful. */

  for (i = 0; i < tinn; i++)
    for (j = 0; j < tinn; j++)
      c[i][j] = 0;

  for (itr = 0; itr < ntriangles; itr++) {
    for (j = 0; j < 3; j++) {	// Color The Vertices. 
      k = triangles[itr].vert[j];
      x = points[k].x;
      y = points[k].y;
      if (x < 0 || x >= tinn || y < 0 || y >= tinn)
	cout << "ERROR: illegal (x,y)=(" << x << "," << y << ") in Write_Pgm\n";
      else
	c[x][y] = 0;
    }
    Pairindex thispair;
    thispair = triangles[itr].l.head;
    while (thispair != eolist) {	// Color The Points In The Triangles. 
      if (thispair < 0)
	perror2 ("Negative thispair in Write_Pgm.");
      k = dotpairs[thispair].ptid;
      thispair = dotpairs[thispair].next;
      x = points[k].x;
      y = points[k].y;
      if (x < 0 || x >= tinn || y < 0 || y >= tinn)
	cout << "ERROR: illegal (x,y)=(" << x << "," << y << ") in Write_Pgm\n";
      else
	c[x][y] = itr;
    }
  }

  ofstream tout ("t.pgm");
  if (!tout)
    perror2 ("Can't open t.pgm file for writing.");
  tout << "P2\n#t.pgm, written by tin.cc\n" << tinn << ' ' << tinn
       << endl << ntriangles << endl;

  k = 0;			// Counter For When To Insert A Newline. 
  for (i = 0; i < tinn; i++)
    for (j = 0; j < tinn; j++) {
      tout << setw (8) << c[i][j];
      if (((++k) & 7) == 0)
	tout << endl;
    }
  if ((k & 7) != 0)
    tout << endl;
}


/* WRITE_EDGES Write the triangles' edges to a file, with each edge:

x1 y1 z1 x2 y2 z2 <blank line>

Altho each edge (except on the outside) occurs in 2 triangles, write
it only once.  Do this by writing an edge iff the adjacent triangle
has a smaller number, which includes -1 for the outside. */

void Write_Edges (const string filename)
{
  int     itr;
  int     j;
  int     a, b;
  ofstream to (filename.c_str());

  if (!to)
    perror2 ("Can't open edges file for writing.");

  to << "# Edges in the triangulation:" << endl
     << "# x0 y0 z0" << endl
     << "# x1 y1 z1" << endl << "# (blank line)" << endl << "# ..." << endl;

  for (itr = 0; itr < ntriangles; itr++)
    for (j = 0; j < 3; j++) {
      a = triangles[itr].vert[j];
      b = triangles[itr].vert[plus1[j]];
      if (itr > triangles[itr].adj[minus1[j]])
	to << setw (5) << points[a].x << setw (5) << points[a].
	  y << setw (6) << points[a].z << endl << setw (5) << points[b].
	  x << setw (5) << points[b].y << setw (6) << points[b].
	  z << endl << endl;
    }
}



/* WRITE_BIN_TRIANGLES Write the triangles in binary form to a file.
   Each triangle is:

   x1 y1 z1 x2 y2 z2 x3 y3 z3

   Each coord is a 2 byte int.

*/

void Write_Bin_Triangles (const string filename)
{
  ofstream to (filename.c_str());

  if (!to)
    perror2 ("Can't open triangles file for writing.");

  for (int itr = 0; itr < ntriangles; itr++) {
    for (int j = 0; j < 3; j++) {
      int    &a = triangles[itr].vert[j];
      to.write (reinterpret_cast < char *>(&points[a]), 6);
    }
  }

}

/* WRITE_TRIANGLES Write the triangles to a file, with each triangle:

x1 y1 z1 x2 y2 z2 x3 y3 z3 x1 y1 z1 <blank line> */

// Don't write triangles now.

void Write_Triangles (const string filename)
{
  return;
  int     itr;
  int     j;
  int     a, b;
  ofstream to (filename.c_str());

  if (!to)
    perror2 ("Can't open triangles file for writing.");

  to << "# Triangles in the triangulation:" << endl
     << "# x0 y0 z0" << endl
     << "# x1 y1 z1" << endl
     << "# x2 y2 z2" << endl
     << "# x0 y0 z0" << endl << "# (blank line)" << endl << "# ..." << endl;

  for (itr = 0; itr < ntriangles; itr++) {
    for (j = 0; j < 3; j++) {
      a = triangles[itr].vert[j];
      to << setw (5) << points[a].x << setw (5) << points[a].y
	 << setw (6) << points[a].z << endl;
    }
    a = triangles[itr].vert[0];
    to << setw (5) << points[a].x << setw (5) << points[a].y
       << setw (6) << points[a].z << endl << endl;
  }
}




// WRITE_TRIANGLE_CENTERS Write the triangles' centers to a file in
// the form of gnuplot set label commands.  For triangle #7:

// set label 8 "7" x y z


void Write_Triangle_Centers (const string filename)
{
  int     itr;
  int     b;
  float   a;
  ofstream to (filename.c_str());

  if (!to)
    perror2 ("Can't open triangle centers file for writing.");

  to << "# tin triangle centers.   do this in gnuplot:" << endl
     << "# load \"centers.all\"" << endl;

  for (itr = 0; itr < ntriangles; itr++) {
    to << "set label " << setw (7) << itr +
      1 << ' ' << '"' << itr << '"';
    a = (points[triangles[itr].vert[0]].x + points[triangles[itr].vert[1]].x +
	 points[triangles[itr].vert[2]].x) * 0.33333;
    to << " at " << setw (7) << a;
    a = (points[triangles[itr].vert[0]].y + points[triangles[itr].vert[1]].y +
	 points[triangles[itr].vert[2]].y) * 0.33333;
    to << ',' << setw (7) << a;
    a = (points[triangles[itr].vert[0]].z + points[triangles[itr].vert[1]].z +
	 points[triangles[itr].vert[2]].z) * 0.33333;
    to << ',' << setw (7) << a << " center\n";
  }
}


// WRITE_PLOIFILE Write an input file for gnuplot, to read and plot
// the edges and triangle names.  This needs the names of the edge and
// center files, to reference them in the plotfile.  If centerfilename
// is an empty string, then don't plot the triangle names.

// Don't write plotsfiles now.

void Write_Plotfile(const string plotfilename, const string centerfilename,
		    const string edgefilename)
{
  return;

  ofstream to (plotfilename.c_str());
  if (!to)
    perror2 ("Can't open plot file for writing.");
  to << "#!/usr/bin/gnuplot\n"
     << "set nokey\n";
  if (centerfilename.size() > 0) 
    to << "load \"" << centerfilename << "\"\n";
  to << "plot \"" << edgefilename << "\" with lines" << endl;
  to << "pause -1"; 
}


void Print_Stats ()
{
  Print_Time2 ("Cumulative processing (not i/o)");
  Print_Time ("Last few triangle splittings");
  cout << "[Tin finished.]\n"
       << "# rows of input points (envar TINN) =                " << tinn << '\n'
       << "# input points =                                     " << npoints << '\n'
       << "Max # splits allocated (envar TSPLITS) input as =    " << tsplits << '\n'
       << "# splits =                                           " << nsplits << '\n'
       << "# output points =                                    " << (nsplits +
								      4) << '\n' <<
    "# output triangles =                                 " << ntriangles << '\n'
       << "# output edges =                                     " << (nsplits +
								      ntriangles +
								      3) << '\n' 
       << "Max error =                                             " << max_err_over_whole_TIN << '\n'
       <<  "When are quadrilateral diagonals flipped?            ";

  switch (tflip) {
  case 0:
    cout << "never";
    break;
  case 1:
    cout << "Delauney";
    break;
  case 2:
    cout << "min weight";
    break;
  case 3:
    cout << "minmax error";
    break;
  case 4:
    cout << "mix of minmax and Delauney";
    break;
  default:
    cout << "illegal value of tflip";
  }
  cout << endl;

  cout << "# times quadrilateral diagonal flipped =             " << nflips << '\n'
       << "Hi-water size for edge flip queue =                  " << hiflipq << '\n'
       << "# times elements pushed onto errs heap =             " <<
    n_elts_pushed_onto_errs_heap << '\n' <<
    "# times elements popped from errs =                  " <<
    n_elts_popped_from_heap << '\n' <<
    "# times elements deleted from errs =                 " <<
    n_elts_deleted_from_heap << '\n' <<
    "# times elements swapped when pushing =              " <<
    n_elts_pushed_onto_errs_heap_swap << '\n' <<
    "# times elements swapped when popping =              " <<
    n_swaps_while_popping_heap << '\n' <<
    "Hi-water size for heap =                             " << max_ever_heap_size
       << '\n';
  //  Write_Bin_Triangles ("trianglesb_all.dat");
  if (twritemore) {
    Write_Triangles ("trianglesa_all.dat");
    Write_Edges ("edges_all.dat");
    Write_Triangle_Centers ("centers_all.dat");
    Write_Plotfile("plot.gnu", "centers_all.dat", "edges_all.dat");
    Write_Pgm ();
  }
  Print_Time ("Writing results");
}


// HEAP_ELT_COMPARE Call by qsort called by heap push.  This is
// defined so as to sort in DECREASING order.  If two heap elements
// have the same error, then don't worry about order.

int Heap_Elt_Compare (const void *a, const void *b)
{
  int     d = (int) (triangles[((const Heap_elt *) b)->tri].maxe -
		     triangles[((const Heap_elt *) a)->tri].maxe);
  return d;
}




/** 
    MAKE_FIRST_2_TRIANGLES  This assumes that the input heights are in a square array
    with y varying faster than x. 


    ...........
    .2      1..
    .       .2.
    .  T1  .  .
    .     .   .
    .    .    .
    .   .     .
    .  .  T0  .
    .0.       .
    ..0      1.
    ...........
*/


void Make_First_2_Triangles ()
{
  Triangle *t0, *t1;		/* Pointers to the triangles, which are really in
				   array 'triangles'. */
  int     i, t;


  if (max_allowable_num_triangles < 2) {
    cout << "max_allowable_num_triangles=" << max_allowable_num_triangles 
	 << " is ridiculously small.\n";
    exit (1);
  }
  t0 = &triangles[0];
  t1 = &triangles[1];

  t0->vert[0] = 0;
  t0->vert[1] = tinn * (tinn - 1);
  t0->vert[2] = tinn * tinn - 1;
  t1->vert[0] = 0;
  t1->vert[1] = tinn * tinn - 1;
  t1->vert[2] = tinn - 1;
  t0->adj[0] = -1;
  t0->adj[1] = 1;
  t0->adj[2] = -1;
  t1->adj[0] = -1;
  t1->adj[1] = -1;
  t1->adj[2] = 0;

  /* 
     Allocate the points to the 2 triangles based on their indices since that
     determines their X and Y coordinates.  The vertices of a triangle are not in
     any triangle. */

  for (i = 0; i < 2; i++)
    triangles[i].l = eolist;

  for (i = 0; i < npoints; i++)
    if (i == 0 || i == tinn - 1 || i == tinn * tinn - 1
	|| i == tinn * (tinn - 1));
    else if (points[i].x > points[i].y) {
      triangles[0].l.push (i);
    } else {
      triangles[1].l.push (i);
    }

  ntriangles = 2;
  Calc_Triangle_Error (triangles[0]);
  errs->push (0);
  Calc_Triangle_Error (triangles[1]);
  errs->push (1);
  Print_Triangle (0);
  Print_Triangle (1);

  Print_Time ("make_1st_2_triangles");
}



/**

SPLIT_TRIANGLE Split the worst triangle, calculate the new
triangles' equations etc, and update max_err_over_whole_TIN etc.
How the 3 new triangles' vertices correspond to the old triangle,
and the signs of the new edge equations. Return 0 if this was
successful, or 1 if no triangle could be split.  This might be
because no triangle has any error.


                                2
                                .
                               ...
                              .2.2.
                             .  .  .
                            .   .   .
                           .    .    .
                        1 .    -.+    . 0
                         .      .      .
new triangle #ntriangles.      1.0      .       new triangle #worst_triangle
                       .      ....       .
                      .    +..  2 ..-     .
                     .    ..-      +..     .
                    .   ..            ..    .
                   . 0..                ..   .
                  . ..                    ..1 .
                 ... 0                     1.. .
                .................................
               0                2                1
                     new triangle #ntriangles+1 */


int Split_Triangle ()
{
  int     i, j, k;
  int     newt;		// new triangle 
  Triangle *thist;
  float   line[3][3];		/* 2D equations of the 3 new edges: line[i][0]*x +
				   line[i][1]*y + line[i][2] = 0 */
  int     newvert;		/* ID of worst point of worst triangle; the new
				   vertex.  */
  Point  *thispt;		// Address Of Coords Of New Point. 
  int     nnewp[3];		// Number Of Points In Each New Triangle. 
  int     onedge[3];		// 1=> the new point is on edge #i of the triangle. 
  float   d;			// Distance Of Current Point From Line. 

  if (tprint > 0)
    cout << "Trying to split the worst triangle.\n";

  if (worst_triangle < 0 || max_err_over_whole_TIN == 0.0
      || triangles[worst_triangle].worst_pt < 0) {
    cout << "No triangle splittable.\n";
    return 1;
  }
  thist = &triangles[worst_triangle];

  // Allocate 2 New Triangles And Split The Worst Triangle. 

  if (ntriangles + 2 > max_allowable_num_triangles) {
    cout << "ERROR: max_allowable_num_triangles=" << max_allowable_num_triangles 
	 << " is too small.\n";
    exit (1);
  }
  newvert = thist->worst_pt;
  if (tprint != 0)
    cout << "Splitting triangle " << worst_triangle <<
      ", with max_err_over_whole_TIN " << max_err_over_whole_TIN <<
      ", at point " << newvert << '\n';

  thispt = &points[newvert];

  /* 
     Find 2D equations of the 3 new edges.  Must do before updating the old
     triangle.  This doesn't overflow because, altho points is short int, line is
     float. */

  for (i = 0; i < 3; i++) {
    line[i][0] = points[thist->vert[i]].y - thispt->y;
    line[i][1] = -points[thist->vert[i]].x + thispt->x;
    line[i][2] = -thispt->x * line[i][0] - thispt->y * line[i][1];
  }

#ifdef DEBUG
  printf ("Line equations of new edges:\n");
  for (i = 0; i < 3; i++)
    printf ("%d: %.3f*x%+.3f*y%+.3f=0.\n", i, line[i][0], line[i][1],
	    line[i][2]);
#endif

  // Test Which Side Of Line L Point P Is On. 

#define TESTLP(l,p) (line[l][0]*points[p].x+line[l][1]*points[p].y+line[l][2])


  // Update The Adjacencies Of The Triangles Adjacent To Worst_Triangle. 

  for (i = 0; i < 3; i++) {
    k = triangles[worst_triangle].adj[i];
    if (k >= 0) {
      // Triangle K Is Adjacent.  Update It. 
      switch (i) {
      case 0:
	newt = worst_triangle;
	break;
      case 1:
	newt = ntriangles;
	break;
      case 2:
	newt = ntriangles + 1;
	break;
      default:
	perror2 ("Shouldn't have hit default in switch in Split_Triangle.");
	exit (1);
      }
      for (j = 0; j < 3; j++)
	if (triangles[k].adj[j] == worst_triangle) {
	  triangles[k].adj[j] = newt;
#ifdef DEBUG
	  printf ("Updating adjacency of triangle %d from %d to %d.\n", k,
		  worst_triangle, newt);
#endif
	  goto foundadj2;
	}
      cout << "ERROR: In Split_Triangle, triangle #" << worst_triangle << " is adjacent to triangle #" << k << ", but the reverse is not true.\n";
    foundadj2:;
    }
  }

  // Create The 3 New Triangles.  One Is The Old Triangle, Updated. 

  triangles[ntriangles].vert[0] = thist->vert[0];
  triangles[ntriangles].vert[1] = newvert;
  triangles[ntriangles].vert[2] = thist->vert[2];

  triangles[ntriangles + 1].vert[0] = thist->vert[0];
  triangles[ntriangles + 1].vert[1] = thist->vert[1];
  triangles[ntriangles + 1].vert[2] = newvert;

  thist->vert[0] = newvert;	// Do This After The Previous Two Triangles. 
  // thist->vert[1], [2] are the same. 

  triangles[ntriangles].adj[0] = worst_triangle;
  triangles[ntriangles].adj[1] = thist->adj[1];
  triangles[ntriangles].adj[2] = ntriangles + 1;

  triangles[ntriangles + 1].adj[0] = worst_triangle;
  triangles[ntriangles + 1].adj[1] = ntriangles;
  triangles[ntriangles + 1].adj[2] = thist->adj[2];

  // thist->adj[0] is the same. 
  thist->adj[1] = ntriangles;
  thist->adj[2] = ntriangles + 1;

  if (fabs (TESTLP (0, thist->vert[1])) < TOLERANCE)
    onedge[2] = 1;
  else
    onedge[2] = 0;
  if (fabs (TESTLP (0, thist->vert[2])) < TOLERANCE)
    onedge[1] = 1;
  else
    onedge[1] = 0;
  if (fabs (TESTLP (1, thist->vert[2])) < TOLERANCE)
    onedge[0] = 1;
  else
    onedge[0] = 0;

  for (i = 0; i < 3; i++)
    if (onedge[i] > 0)
      if (tprint != 0)
	cout << "New vertex is on edge #" << i << " of triangle.\n";

  /* 
     Allocate each point from the old triangle, except the new vertex,
     to one of the three new triangles.  Points that are on a new edge
     can be put into either adjacent triangle unless one adjacent
     triangle is degenerate.  In that case, put the point in the other
     triangle.  The problem that this avoids is as follows.  A
     degenerate triangle is probably a vertical plane; I can't
     calculate a plane equation, so I can't calculate errors of points
     from it (or rather they are infinite). */

  // In terms of the former version: 0: Worst, 1: Ntriangles, 2: Ntriangles+1 

  Linked_list list0, list1, list2;
  list0 = list1 = list2 = eolist;

  while (thist->l.head != eolist) {
    n_split_tri_list_traverses++;
    j = thist->l.pop ();
    if (j < 0)
      perror2 ("Illegal point popped in Split_Triangle.");
    if (j == newvert);
    else {
      d = TESTLP (0, j);
      if (d > TOLERANCE || (d >= 0 && onedge[1] == 0)) {
	d = TESTLP (2, j);
	if (d > TOLERANCE || (d >= 0 && onedge[0] == 0))
	  list0.push (j);
	else
	  list1.push (j);
      } else {
	d = TESTLP (1, j);
	if (d > TOLERANCE || (d >= 0 && onedge[2] == 0))
	  list2.push (j);
	else
	  list0.push (j);
      }
    }
  }

  triangles[worst_triangle].l = list0;
  triangles[ntriangles].l = list1;
  triangles[ntriangles + 1].l = list2;

  // already deleted when popped: errs->del(triangles[worst_triangle].heap_loc);

  Calc_Triangle_Error (triangles[worst_triangle]);
  errs->push (worst_triangle);
  Calc_Triangle_Error (triangles[ntriangles]);
  errs->push (ntriangles);
  Calc_Triangle_Error (triangles[ntriangles + 1]);
  errs->push (ntriangles+1);

  /* 
     Add edges between the three new triangles and the adjacent old triangles to
     the list of quadrilateral diagonals to be tested to flipping. */

  Add_to_Flipq (worst_triangle, 0);
  Add_to_Flipq (ntriangles, 1);
  Add_to_Flipq (ntriangles + 1, 2);

  ntriangles += 2;

  Print_Triangle (worst_triangle);
  Print_Triangle (ntriangles - 2);
  Print_Triangle (ntriangles - 1);

  /* Check triangles last to first, else if there are 2 null triangles, then
     deleting the first will invalidate the deletable_triangle array. */

  Flag_Null_Triangle_For_Deletion (ntriangles - 1);
  Flag_Null_Triangle_For_Deletion (ntriangles - 2);
  Flag_Null_Triangle_For_Deletion (worst_triangle);

  Delete_Flagged_Triangles ();

  return 0;

}




/**
   PROC_FLIPQ  Test diagonals in flipq and maybe flip them.
*/


void Proc_Flipq ()
{
  int     k, ixy, n, t, ta;
  int     a[2];		/* which adj entry in each triangle points to the
			   other. */
  COORDT  oldedge[2][3][2];	/* The edge vectors of the two current
				   triangles: oldedge[which
				   triangle][which edge of it][x or
				   y]. */

  COORDT  newedge[2][3][2];	// Edge Vectors Of Proposed New Triangles. 
  COORDT  line[3];		/* Equation of new diagonal:
				   line[0]*x+line[1]*y+line[2]=0. */
  Point  *thisp;		// Pointer To Current Point, In Array
				// 'Points'.
  int    *(oldp[2]);		/* Copy of pointers to p arrays for
				   old triangles, while I'm creating
				   new p arrays. */
  float oldmaxcos, newmaxcos;


  if (tprint>=1) {
    cout << "Starting Proc_Flipq, nflipq=" << nflipq << ", flipq=" << endl;
    for (int iflip=0; iflip < nflipq; iflip++)
      cout << flipq[iflip][0] << ' ' << flipq[iflip][1] << endl;
  }

  /* 
     Repeat for each candidate diagonal.  WARNING: The body of this
     loop may add diagonals to the q, and so increase nflipq. */

  for (int iflip = 0; iflip < nflipq; iflip++) {
    int flippossible;     // Is this flip even possible?
    int     doflip=0;		// Do we flip this diagonal? 1=yes 0=no.
    int    t0 = flipq[iflip][0];	// IDs of the 2 adjacent triangles.
    int    t1 = flipq[iflip][1];

    if (t0<0 || t0>=ntriangles || t1<0 || t1>=ntriangles) {
      cout << "Error in Proc_Flipq: illegal flipq[" << iflip << "] = (" << t0 << ',' << t1 << ')' << endl;
      cout << "nflipq=" << nflipq << ", flipq=" << endl;
      for (int iflip=0; iflip < nflipq; iflip++)
	cout << flipq[iflip][0] << ' ' << flipq[iflip][1] << endl;
      exit(1);
    }

    // Pointers to the two old triangles.
    Triangle *thist[2] = { &triangles[t0], &triangles[t1] };

    Triangle newt[2];           // New triangles.

    newt[0].l = eolist;
    newt[1].l = eolist;

    // For each of the 2 triangles, find which adj points to the other. 

    // For each of the diagonal's adjacent triangles...
    for (int itr = 0; itr < 2; itr++) {
      // Check all of this triangle's adjacent triangles to find the match.
      for (int j = 0; j < 3; j++)	
	if (thist[itr]->adj[j] == flipq[iflip][1 - itr]) {
	  a[itr] = j;
	  goto foundflip;
	}
      cout << "ERROR: in Proc_Flipq, cannot find a[itr=" << itr 
	   << "]= " << a[itr] << " for flipq[iflip=" << iflip << "][1-itr=" 
	   << (1-itr) << "]=" << flipq[iflip][1-itr] << endl;
      cout << "nflipq=" << nflipq << ", flipq=" << endl;
      for (int iflip=0; iflip < nflipq; iflip++)
	cout << flipq[iflip][0] << ' ' << flipq[iflip][1] << endl;
      Write_Pgm();
      Write_Triangles("trianglesa.all");
      exit (1);
    foundflip:;
    }

    // Find the edge vectors of each old triangle.  Make each edge be
    // opposite the corresponding vertex.

    for (int itr = 0; itr < 2; itr++)
      for (int j = 0; j < 3; j++) {
	// ids of 2 current vertices defining an edge. 
	const int &v1 = thist[itr]->vert[plus1[j]];
	const int &v0 = thist[itr]->vert[minus1[j]];
	oldedge[itr][j][0] = points[v1].x - points[v0].x;
	oldedge[itr][j][1] = points[v1].y - points[v0].y;
      }

    // Load  the vertices of the hypothetical new triangles.

    newt[0].vert[0] = thist[0]->vert[minus1[a[0]]];
    newt[0].vert[1] = thist[0]->vert[a[0]];
    newt[0].vert[2] = thist[1]->vert[a[1]];
    newt[1].vert[0] = thist[0]->vert[a[0]];
    newt[1].vert[1] = thist[0]->vert[plus1[a[0]]];
    newt[1].vert[2] = thist[1]->vert[a[1]];

    // Calculate edge vectors of the hypothetical new triangles.

    for (int itr = 0; itr < 2; itr++)
      for (int j = 0; j < 3; j++) {
	// ids of 2 current vertices defining an edge. 
	const int &v1 = newt[itr].vert[plus1[j]];
	const int &v0 = newt[itr].vert[minus1[j]];
	newedge[itr][j][0] = points[v1].x - points[v0].x;
	newedge[itr][j][1] = points[v1].y - points[v0].y;
      }

    flippossible = Is_Flip_Possible(newedge);

    // Now, handle the different flip criteria.

    if (flippossible) {
      Find_Angle_Cosines(oldedge, newedge, oldmaxcos, newmaxcos);

      switch (tflip) {
      case 0:
	break;		// never flip
      case 1:		// Delauney
	{
	  // Flip diagonals if necessary to maintain a Delauney
	  // triangulation, i.e., to maximize the minimum angle in the
	  // 2 triangles.  If the min new angle is larger than the min
	  // old angle, and the new triangles have positive angles,
	  // then flip.

	  if (newmaxcos < oldmaxcos)
	    doflip = 1;
	  break;
	}

      case 2:		// min weight
	{
	  // Squared lengths of the old and new diagonals.
	  const float oldsqlen 
	    = Square(oldedge[0][a[0]][0])+Square(oldedge[0][a[0]][1]);
	  const float newsqlen 
	    = Square(newedge[0][0][0])+Square(newedge[0][0][1]);
	  if (tprint>1) 
	    cout << "oldsqlen=" << oldsqlen << ", newsqlen=" << newsqlen << endl;
	  if (newsqlen < oldsqlen)
	    doflip = 1;
	  break;
	}

      // mix of minmax and Delauney: If an input triangle is too thin, then
      // use the Delauney criterion.  Else use the minmax criterion.
      // 0.99 is a fudge factor.

      case 4:  
	{ 
	  if (oldmaxcos > 0.99 && oldmaxcos > newmaxcos)
	    {
	      doflip = 1;
	      if (tprint>1) 
		cout << "Flipping because of Delauney" << endl;
	      break;
	    }
	  // No break here!  Fall thru from case 4 to case 3.
	}

      case 3:		// min max error
	{
	  const float olderr=max(thist[0]->maxe, thist[1]->maxe);
	  int     j;

	  // Find the equation of the new diagonal. The positive side is
	  // triangle t1; the negative side t0.

	  line[0] = points[thist[0]->vert[2]].y - points[thist[0]->vert[1]].y;
	  line[1] = points[thist[0]->vert[1]].x - points[thist[0]->vert[2]].x;
	  line[2] =
	    -points[thist[0]->vert[2]].x * line[0] -
	    points[thist[0]->vert[2]].y * line[1];

	  // Copy each point from the two old triangles to the proper new
	  // triangle.  

	  for (int itr = 0; itr < 2; itr++) {
	    Pairindex p = thist[itr]->l.head;
	    while (p!= eolist) {
	      j = dotpairs[p].ptid;
	      p = dotpairs[p].next;
	      Point &pt = points[j];
	      if (line[0] * pt.x + line[1] * pt.y + line[2] > 0) {
		newt[1].l.push (j);
	      } else {
		newt[0].l.push (j);
	      }
	    }
	  }

	  Calc_Triangle_Error(newt[0]);
	  Calc_Triangle_Error(newt[1]);

	  // Free the new triangle point lists.

	  for (int itr = 0; itr < 2; itr++) {
	    while (newt[itr].l.head != eolist) {
	      newt[itr].l.pop ();
	    }
	  }

	  const float newerr = max(newt[0].maxe, newt[1].maxe);

	  if (tprint>0) 
	    cout << "Old triangle errors=" << thist[0]->maxe << ',' << thist[1]->maxe 
		 << ", new triangle errors= " << newt[0].maxe << ',' << newt[1].maxe 
		 << endl; 
	  doflip = (newerr < olderr);

	  COORDT olddet[2] = { Det3(oldedge[0]), Det3(oldedge[1]) };

	  if (tprint>0)
	    cout << "Old dets= " << olddet[0] << ", " << olddet[1] << endl;

	  if (olddet[0]==0 || olddet[1]==0) {
	    if (tprint>0) 
	      cout << "Forcing flip since one triangle is degenerate." << endl; 
	    doflip=1;
	  }
	  break;
	}

      default:
	cout << "Illegal tflip in proc_flipq " << tflip << endl;
	exit (1);
      }

    }

    if (doflip) {
      Flip_Diagonal(t0, t1, newt, line, a, iflip);
    } else if (tprint > 0)
      cout << "Not flipping triangle #" << t0 << " and triangle #" << t1 << endl;

  }				// End of processing this flipq entry. 
  nflipq = 0;
}




// IS_FLIP_POSSIBLE Find if flipping would produce an illegal triangle,
// i.e., is the quadrilateral concave?

int Is_Flip_Possible(COORDT newedge[2][3][2])
{
  COORDT  det;		// Determinant of a pair of edges in one of the
                        // new triangles.
  COORDT  mindet;	// Smallest determinant.
  mindet = 9999;	// Precise value unimportant; need to know only if
                        // negative.
  int flippossible;

  // Find the determinants of 3 pairs of edges for each of the 2 new
  // triangles, in order to check whether a new triangle is oriented
  // backwards.  This will happen if the proposed new diagonal lies outside
  // the triangles.  (You might want to check the adjacent new angles are
  // acute, but this is too strict.) It's sufficient to do the 2 pairs
  // representing the new angles, but determining which two is a little more
  // complicated.

  for (int itr = 0; itr < 2; itr++) {
    det = Det2D (newedge[itr][0], newedge[itr][1]);
    Mineq (mindet, det);
    det = Det2D (newedge[itr][1], newedge[itr][2]);
    Mineq (mindet, det);
    det = Det2D (newedge[itr][2], newedge[itr][0]);
    Mineq (mindet, det);
  }
  flippossible= (mindet>0);
  if (tprint >0 && !flippossible) 
    cout << "This flip is impossible." << endl;
  return flippossible;
}





// FIND_ANGLE_COSINES Find the angle cosines, for use in Delauney
// flipping. The cosines are proxies for the angles themselves, but save
// trig evaluations.  The minus is needed because each edge vector is coming
// from a different vertex.

void Find_Angle_Cosines(COORDT oldedge[2][3][2], COORDT newedge[2][3][2],
			float &oldmaxcos, float &newmaxcos) 
{
  float   oldcos[2][3], newcos[2][3];	/* cosines of the angles of
					   the existing, and proposed
					   new, triangles. */
  oldmaxcos = -100.0;
  for (int itr = 0; itr < 2; itr++)
    for (int j = 0; j < 3; j++) {
      oldcos[itr][j] =
	-Dot2D (oldedge[itr][minus1[j]], oldedge[itr][plus1[j]]) /
	sqrt (Dot2D
	      (oldedge[itr][minus1[j]],
	       oldedge[itr][minus1[j]]) *
	      Dot2D (oldedge[itr][plus1[j]],
		     oldedge[itr][plus1[j]]));
      if (oldcos[itr][j] > oldmaxcos)
	oldmaxcos = oldcos[itr][j];
    }

  newmaxcos = -100.;

  for (int itr = 0; itr < 2; itr++) {
    for (int j = 0; j < 3; j++) {	// Find The Angle Cosines 
      newcos[itr][j] =
	-Dot2D (newedge[itr][j], newedge[itr][plus1[j]]) /
	sqrt (Dot2D (newedge[itr][j], newedge[itr][j]) *
	      Dot2D (newedge[itr][plus1[j]],
		     newedge[itr][plus1[j]]));
      if (newcos[itr][j] > newmaxcos)
	newmaxcos = newcos[itr][j];
    }
  }
}




void Flip_Diagonal(int t0, int t1, Triangle *newt, COORDT *line, int *a, int iflip)
{
  Triangle *thist[2] = { &triangles[t0], &triangles[t1] };
  Point  *thisp;		// Pointer To Current Point, In Array 'Points'. 

  if (tprint != 0)
    cout << "Flip #" << nflips << ": Flipping triangle #" << t0
	 << " (heap_loc=" << triangles[t0].heap_loc
	 << ") and triangle #" << t1 << " (heap_loc="
	 << triangles[t1].heap_loc << ")" << endl;
  nflips++;

  if (triangles[t0].heap_loc< -1 || triangles[t0].heap_loc==0 
      || triangles[t0].heap_loc > errs->heap_size) {
    cout << "ERROR:  triangles[t0=" << t0 << "].heap_loc=" 
	 << triangles[t0].heap_loc << " is illegal.   errs->heap_size=" 
	 << errs->heap_size << endl;
    exit(1);
  }

  if (triangles[t1].heap_loc< -1 || triangles[t1].heap_loc==0 
      || triangles[t1].heap_loc > errs->heap_size) {
    cout << "ERROR:  triangles[t1=" << t1 << "].heap_loc=" 
	 << triangles[t1].heap_loc << " is illegal.   errs->heap_size=" 
	 << errs->heap_size << endl;
    exit(1);
  }

  /**
	v[a[0]] means thist[0]->vert[a[0]] etc

	OLD                                   NEW

	v[minus1[a[1]]]            2
	v[a[1]]......................                ......................1
	       .                   ..v[plus1[a[0]]]  ..                   .
	       .                  . .               2. .                  .
	       .                 .  .                .  .                 .
	       .                .   .                .   .                .
	       .               .    .                .    .               .
	       .      T1      .     .                .     .       T1     .
	       .             .      .                .      .             .
	       .            .       .                .       .            .
	       .           .        .                .        .           .
	       .          .         .                .         .          .
	       .         .          .                .          .         .
	       .        .           .                .           .        .
	       .       .            .                .            .       .
	       .      .             .                .             .      .
	       .     .     T0       .                .       T0     .     .
	       .    .               .                .               .    .
	       .   .                .                .                .   .
	       .  .                 .                .                 .  .
	       . .                  .                .                  . .
	       ..                   .                .                   ..
 v[plus1[a[1]]]......................                ...................... 0
		v[minus1[a[0]]]     v[a[0]]          0                   1


  */

  /* The following errors mean that my internal data structure is invalid. 
   */

  if (thist[0]->vert[plus1[a[0]]] != thist[1]->vert[minus1[a[1]]]) {
    cout 
      << "ERROR: in Proc_Flipq, thist[0]->vert[plus1[a[0]]]!=thist[1]->vert[minus1[a[1]]]\n";
    exit (1);
  }
  if (thist[0]->vert[minus1[a[0]]] != thist[1]->vert[plus1[a[1]]]) {
    cout << "ERROR: in Proc_Flipq, thist[0]->vert[minus1[a[0]]]!=thist[1]->vert[plus1[a[1]]]\n";
    exit (1);
  }
  /* 
     Calculate the new adj into a new array, then copy vert and adj back. */

  newt[0].adj[0] = t1;
  newt[0].adj[1] = thist[1]->adj[minus1[a[1]]];
  newt[0].adj[2] = thist[0]->adj[plus1[a[0]]];
  newt[1].adj[0] = thist[1]->adj[plus1[a[1]]];
  newt[1].adj[1] = t0;
  newt[1].adj[2] = thist[0]->adj[minus1[a[0]]];

  /* 
     Flipping the diagonal may change flipq entries referring to these
     triangles.  Update them. */

  for (int i = iflip + 1; i < nflipq; i++)
    if (flipq[i][0] == t1 && flipq[i][1] == thist[1]->adj[minus1[a[1]]]) {
      flipq[i][0] = t0;
      if (tprint != 0)
	cout << "Updating flipq[" << i << "][0] from " << t1 << " to " << t0 
	     << " because of diagonal flip.\n";
    } else if (flipq[i][1] == t1
	       && flipq[i][0] == thist[1]->adj[minus1[a[1]]]) {
      flipq[i][1] = t0;
      if (tprint != 0)
	cout << "Updating flipq[" << i << "][1] from " << t1 << " to " << t0 
	     << " because of diagonal flip.\n";
    } else if (flipq[i][0] == t0
	       && flipq[i][1] == thist[0]->adj[minus1[a[0]]]) {
      flipq[i][0] = t1;
      if (tprint != 0)
	cout << "Updating flipq[" << i << "][0] from " << t0 << " to " << t1
	     << " because of diagonal flip.\n";
    } else if (flipq[i][1] == t0
	       && flipq[i][0] == thist[0]->adj[minus1[a[0]]]) {
      flipq[i][1] = t1;
      if (tprint != 0)
	cout << "Updating flipq[" << i << "][1] from " << t0 << " to " << t1
	     << " because of diagonal flip.\n";
    }
  // Now Update The Vert And Adj Entries. 

  for (int itr = 0; itr < 2; itr++)
    for (int j = 0; j < 3; j++) {
      thist[itr]->vert[j] = newt[itr].vert[j];
      thist[itr]->adj[j] = newt[itr].adj[j];
    }

  /* 
     Update the adjacencies of the two new triangles with their neighbors. 
     Don't update their adjacencies with each other; that's the j==itr
     test.  It uses the fact that adj[0] of t0 is t1

     etc. */

  for (int itr = 0; itr < 2; itr++) {
    int t = flipq[iflip][itr];
    for (int j = 0; j < 3; j++) {
      if (j == itr)
	continue;
      int ta = thist[itr]->adj[j];
      if (ta < 0)
	continue;
      for (int k = 0; k < 3; k++) {
	if (triangles[ta].adj[k] == t0 || triangles[ta].adj[k] == t1) {
	  triangles[ta].adj[k] = t;
	  goto found1adj;
	}
      }
      cout << "ERROR: When updating 1 adjacency, triangles[" << t << "].adj[" << j 
	   << "]=" << ta << " is not matched by the other triangle.\n";
      exit (1);
    found1adj:;
    }
  }

  /* 
     Find the equation of the new diagonal. The positive side is triangle
     t1; the negative side t0. */

  line[0] = points[thist[0]->vert[2]].y - points[thist[0]->vert[1]].y;
  line[1] = points[thist[0]->vert[1]].x - points[thist[0]->vert[2]].x;
  line[2] =
    -points[thist[0]->vert[2]].x * line[0] -
    points[thist[0]->vert[2]].y * line[1];
#ifdef DEBUG
  printf ("New diagonal is %.3f*x%+.3f*y%+.3f=0.\n", line[0], line[1],
	  line[2]);
#endif

  // Reassign Each Point From The Two Old Triangles To The Proper New
  // Triangle.  

  Linked_list linked_list0, linked_list1;
  int     j;

  linked_list0 = eolist;
  linked_list1 = eolist;

  for (int itr = 0; itr < 2; itr++) {
    while (thist[itr]->l.head != eolist) {
      j = thist[itr]->l.pop ();
      thisp = &points[j];
      if (line[0] * thisp->x + line[1] * thisp->y + line[2] > 0) {
	linked_list1.push (j);
      } else {
	linked_list0.push (j);
      }
    }
  }

  triangles[t0].l = linked_list0;
  triangles[t1].l = linked_list1;

  errs->del (triangles[t0].heap_loc);	// Delete old triangles from errs
  // heap.
  errs->del (triangles[t1].heap_loc);

  Calc_Triangle_Error (triangles[t0]);
  errs->push (t0);
  Calc_Triangle_Error (triangles[t1]);
  errs->push (t1);

  Print_Triangle (t0);
  Print_Triangle (t1);

  /* 
     Add new neighboring edges of the two triangles to flipq.  Do not add
     the diagonal between them. */

  for (int itr = 0; itr < 2; itr++)
    for (int j = 0; j < 3; j++)
      if (thist[itr]->adj[j] != flipq[iflip][1 - itr])
	Add_to_Flipq (flipq[iflip][itr], j);
}





// PRINT_HELP Copy the file tin.help to stdout and exit.

void Print_Help ()
{
  char   *argv[] = { "bincat", "tin.help", NULL };
  execve ("/bin/cat", argv, NULL);
}




// HEAP:PUSH
// Push given triangle onto the heap, if its error > 0.

void Heap::push (const int t)
{
  int father, current;

  if (triangles[t].maxe == 0.0) {
    if (tprint > 0)
      cout << "Not pushing triangle #" << t << " since its err=0." << endl;
    triangles[t].heap_loc = -1;
    return;
  }

  if (heap_size >= alloc_size)
    perror2 ("Heap overflow.");
  current = ++heap_size;
  max_ever_heap_size = max (max_ever_heap_size, heap_size);
  elts[heap_size].tri = t;
  triangles[t].heap_loc = heap_size;
  n_elts_pushed_onto_errs_heap++;
  if (tprint > 0)
    cout << "Pushing triangle #" << t << ", err=" << triangles[t].maxe
	 << ", heap_size=" << heap_size << " now.\n";
  Test_Tri (elts[heap_size].tri, "after pushing message in heap::push");

  while (current > 1)		// Rotate to float the largest elt.
    {
      father = current >> 1;
      if (elts[current].err () <= elts[father].err ())
	break;
      n_elts_pushed_onto_errs_heap_swap++;
      if (tprint > 0)
	cout << "Heap push swapping " << current << ' ' << elts[current] <<
	  " and " << father << ' ' << elts[father] << '\n';
      swap_heap_elt (elts[current], elts[father]);
      current = father;
    }
}


// HEAP::POP

// Pop triangle with largest err, setting global vars worst_triangle and
// max_err_over_whole_TIN to the values popped off.  If the element popped is not a
// valid triangle, then repeat.  If there is no valid element, set -1 and
// 0.0 .  

// Element 0 is not used; the first heap element is 1.  This simplifies programming.

void Heap::pop ()
{
  int c;			// Current elt's id in heap.
  int s;			// First son's id.  Second is s+1, if it exists.
  int l;			// Larger son's id.

  if (heap_size == 0) {
    cout << "Trying to pop an empty err heap.\n";
    worst_triangle = -1;
    max_err_over_whole_TIN = 0.0;
    return;
  }

  max_err_over_whole_TIN = elts[1].err ();
  worst_triangle = elts[1].tri;
  Test_Tri (worst_triangle, "near start of heap::pop");
  n_elts_popped_from_heap++;
  errs->del (1);
  if (tprint > 0)
    cout << "Popping triangle #" << worst_triangle <<
      " with max_err_over_whole_TIN=" << max_err_over_whole_TIN <<
      ", heap_size=" << heap_size << " now, nt=" << ntriangles << ".\n";
}


// HEAP::DEL

// Delete heap element #d, rotating if necessary.

void Heap::del (const int d)
{
  int son, current, i, l;

  if (d < 0) {
    if (tprint > 0)
      cout << "Not deleting err heap element #" << d << endl;
    return;
  }

  i = elts[d].tri;
  swap_heap_elt (elts[d], elts[heap_size]);	// Move the last elt to replace the first
  heap_size--;
  n_elts_deleted_from_heap++;

  if (tprint > 0)
    cout << "Deleting heap element #" << d << ": triangle #" << i
	 << ", giving heap_size=" << heap_size << endl;

  current = d;
  for (;;)			// If the first is not larger than its sons, rotate.
    {
      son = current << 1;
      if (son > heap_size)
	break;		// Don't rotate off the end of the heap.
      if (son == heap_size)	// Is there only one son?
	{
	  if (elts[current].err () >= elts[son].err ())
	    break;
	  l = son;
	} else {
	  if (elts[current].err () >= elts[son].err ()
	      && elts[current].err () >= elts[son + 1].err ())
	    break;
	  l = son;
	  if (elts[son + 1].err () > elts[son].err ())
	    l = son + 1;
	}
      n_swaps_while_popping_heap++;
      if (tprint > 0)
	cout << "Heap pop swapping " << current << ' ' << elts[current] <<
	  " and " << l << ' ' << elts[l] << '\n';
      swap_heap_elt (elts[current], elts[l]);
      current = l;
    }
}





// TEST_TRI: Test whether this is the triangle that we're watching.
// This gives us a chance to break here.

void Test_Tri (const int t, const string msg)
{
  static int tt = -1;

  return;
  if (t == tt) {
    cout << "******* I see triangle #" << t << ":" << msg << '\n';
  }
}



// TEST_PT: Test whether this is the point that we're watching.
// This gives us a chance to break here.

void Test_Pt (const int t, const string msg)
{
  static int tt = -1;

  if (t == tt) {
    cout << "******* I see point #" << t << ":" << msg << '\n';
  }
  return;
}


// WRITE HEAP_ELT

ostream & operator << (ostream & s, Heap_elt e) {
  return s << e.tri;
}



// LINKED_LIST CLASS ROUTINES

void Linked_list::push (int point)	// Prefix a new point to this list
{
  Pairindex i;

  Test_Pt (point, "Pushing onto list.");
  i = freelist.head;
  if (i == eolist)
    perror2 ("Freelist exhausted.");
  freelist.head = dotpairs[i].next;
  dotpairs[i].ptid = point;
  dotpairs[i].next = head;
  head = i;
  this->check ();
}

int Linked_list::pop ()		// Pop the first point off the list.
{
  int i, point;
  if (head == eolist)
    perror2 ("Trying to pop a null list.");
  point = dotpairs[head].ptid;
  Test_Pt (point, "Popping off list.");
  i = dotpairs[head].next;
  dotpairs[head].ptid = -9999;
  dotpairs[head].next = freelist.head;
  freelist.head = head;
  head = i;
  freelist.check ();
  return point;
}




// AREA of a triangle in 2D.

float Area (const Triangle t)
{
  float a0 = points[t.vert[0]].x - points[t.vert[1]].x;
  float a1 = points[t.vert[0]].y - points[t.vert[1]].y;
  float b0 = points[t.vert[0]].x - points[t.vert[2]].x;
  float b1 = points[t.vert[0]].y - points[t.vert[2]].y;

  float area = (a0 * b1 - a1 * b0) * 0.5;

  return area;
}


// COPY_TRIANGLE Copy one triangle to another, given their ids in the
// triangles array.  Also update the heap.

void Copy_Triangle (int to, int from)
{
  if (from < 0 || from >= ntriangles || to < 0 || to >= ntriangles)
    perror2 ("Illegal input to Copy_Triangle.");
  triangles[to] = triangles[from];
  if (triangles[to].heap_loc >= 0)
    errs->elts[triangles[to].heap_loc].tri = to;
}
